use crate::archive::Entry;
use crate::error::ArchiveError;
use crate::read::Archive;
use libarchive3_sys::ffi;
use libc::{c_void, ssize_t};
use std::io;
use std::io::Read;
use std::marker;

pub struct ReadEntry<'a, T: Archive> {
    archive: *mut ffi::Struct_archive,
    entry: *mut ffi::Struct_archive_entry,
    offset: usize,
    _ignored: marker::PhantomData<&'a T>,
}

//impl<'b, T: Archive + 'b> ReadEntry<'b, T> {
//    /// Create a new ReaderEntry from a raw pointer to an `archive_entry` struct
//    ///
//    /// # Safety
//    /// The pointer must own the struct it points to, otherwise it may be freed twice
//    pub unsafe fn new(archive: &'b mut T, handle: *mut ffi::Struct_archive_entry) -> Self {
//        Self { archive, handle }
//    }
//}

impl<'a, T: Archive> Entry for ReadEntry<'a, T> {
    unsafe fn entry(&self) -> *const ffi::Struct_archive_entry {
        self.entry as *const _
    }

    unsafe fn entry_mut(&mut self) -> *mut ffi::Struct_archive_entry {
        self.entry
    }
}

pub struct Entries<'a, T: 'a + Archive> {
    archive: &'a mut T, // entry: &'b mut ReadEntry<'b, T>
                        // entry_handle: *mut ffi::Struct_archive_entry, // entry: ReadEntry<'a, T>
}

// impl Default for ReadEntry {
//     fn default() -> Self {
//         Self {
//             handle: std::ptr::null_mut(),
//         }
//     }
// }

impl<'a, T: 'a + Archive> Entries<'a, T> {
    pub fn new(archive: &'a mut T) -> Self {
        Self { archive }
    }
}

// impl<'a, T: 'a + Archive> EntryFields<'a, T> {
//     fn read_next_header(&'a mut self) -> crate::Result<Option<ReadEntry<'a, T>>> {
//         let mut entry_ptr: *mut ffi::Struct_archive_entry = std::ptr::null_mut();

//         let res = unsafe {
//             ffi::archive_read_next_header(self.archive.handle() as *mut _, &mut entry_ptr)
//         };

//         match res {
//             ffi::ARCHIVE_OK | ffi::ARCHIVE_WARN => Ok(Some(unsafe {
//                 EntryFields::new(self.archive, entry_ptr).into_entry()
//             })),
//             ffi::ARCHIVE_EOF => Ok(None),
//             _ => Err(ArchiveError::Sys(
//                 self.archive.err_code(),
//                 self.archive.err_msg().to_owned(),
//             )),
//         }
//     }
// }

impl<'a, T: Archive + 'a> Iterator for Entries<'a, T> {
    type Item = crate::Result<ReadEntry<'a, T>>;

    fn next(&mut self) -> Option<Self::Item> {
        let mut entry_ptr: *mut ffi::Struct_archive_entry = std::ptr::null_mut();

        let res = unsafe {
            ffi::archive_read_next_header(self.archive.handle() as *mut _, &mut entry_ptr)
        };

        match res {
            ffi::ARCHIVE_OK | ffi::ARCHIVE_WARN => {
                let entry = ReadEntry {
                    archive: unsafe { self.archive.handle_mut() },
                    entry: entry_ptr,
                    offset: 0,
                    _ignored: marker::PhantomData,
                };

                Some(Ok(entry))
            }
            ffi::ARCHIVE_EOF => None,
            _ => Some(Err(ArchiveError::Sys(
                self.archive.err_code(),
                self.archive.err_msg().to_owned(),
            ))),
        }
    }
}

impl<'a, T: Archive> Read for ReadEntry<'a, T> {
    fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
        let res =
            unsafe { ffi::archive_read_data(self.archive, buf as *mut [u8] as *mut _, buf.len()) };

        if res >= 0 {
            let count: usize = res
                .try_into()
                .map_err(|e| io::Error::new(io::ErrorKind::Other, e))?;
            self.offset += count;

            Ok(count)
        } else {
            Err(io::Error::new(io::ErrorKind::Other, "error"))
        }
    }
}
